// SPDX-License-Identifier: Apache-2.0
// Copyright 2023-2024 Jussi Pakkanen

#pragma once

#include <stdint.h>

#define CAPYPDF_VERSION_STR "0.11.0"
#define CAPYPDF_VERSION_MAJOR 0
#define CAPYPDF_VERSION_MINOR 11
#define CAPYPDF_VERSION_MICRO 0

#if defined _WIN32
#ifdef BUILDING_CAPYPDF
#define CAPYPDF_PUBLIC __declspec(dllexport)
#else
#define CAPYPDF_PUBLIC __declspec(dllimport)
#endif
#else
#ifdef BUILDING_CAPYPDF
#define CAPYPDF_PUBLIC __attribute__((visibility("default")))
#else
#define CAPYPDF_PUBLIC
#endif
#endif

#ifdef __cplusplus
#define CAPYPDF_NOEXCEPT noexcept
extern "C" {
#else
#define CAPYPDF_NOEXCEPT
#endif

typedef enum {
    CAPY_FONT_TIMES_ROMAN,
    CAPY_FONT_HELVETICA,
    CAPY_FONT_COURIER,
    CAPY_FONT_SYMBOL,
    CAPY_FONT_TIMES_ROMAN_BOLD,
    CAPY_FONT_HELVETICA_BOLD,
    CAPY_FONT_COURIER_BOLD,
    CAPY_FONT_ZAPF_DINGBATS,
    CAPY_FONT_TIMES_ROMAN_ITALIC,
    CAPY_FONT_HELVETICA_OBLIQUE,
    CAPY_FONT_COURIER_OBLIQUE,
    CAPY_FONT_TIMES_ROMAN_BOLDITALIC,
    CAPY_FONT_HELVETICA_BOLDOBLIQUE,
    CAPY_FONT_COURIER_BOLDOBLIQUE,
} CapyPDF_Builtin_Fonts;

typedef enum {
    CAPY_DEVICE_CS_RGB,
    CAPY_DEVICE_CS_GRAY,
    CAPY_DEVICE_CS_CMYK
} CapyPDF_DeviceColorspace;

typedef enum {
    CAPY_IMAGE_CS_RGB,
    CAPY_IMAGE_CS_GRAY,
    CAPY_IMAGE_CS_CMYK,
    // CAPY_IMAGE_CS_LAB,
} CapyPDF_ImageColorspace;

typedef enum {
    CAPY_RI_RELATIVE_COLORIMETRIC,
    CAPY_RI_ABSOLUTE_COLORIMETRIC,
    CAPY_RI_SATURATION,
    CAPY_RI_PERCEPTUAL
} CapyPDF_Rendering_Intent;

typedef enum {
    CAPY_TEXT_FILL,
    CAPY_TEXT_STROKE,
    CAPY_TEXT_FILL_STROKE,
    CAPY_TEXT_INVISIBLE,
    CAPY_TEXT_FILL_CLIP,
    CAPY_TEXT_STROKE_CLIP,
    CAPY_TEXT_FILL_STROKE_CLIP,
    CAPY_TEXT_CLIP,
} CapyPDF_Text_Mode;

typedef enum {
    CAPY_BM_NORMAL,
    CAPY_BM_MULTIPLY,
    CAPY_BM_SCREEN,
    CAPY_BM_OVERLAY,
    CAPY_BM_DARKEN,
    CAPY_BM_LIGHTEN,
    CAPY_BM_COLORDODGE,
    CAPY_BM_COLORBURN,
    CAPY_BM_HARDLIGHT,
    CAPY_BM_SOFTLIGHT,
    CAPY_BM_DIFFERENCE,
    CAPY_BM_EXCLUSION,
    CAPY_BM_HUE,
    CAPY_BM_SATURATION,
    CAPY_BM_COLOR,
    CAPY_BM_LUMINOSITY,
} CapyPDF_Blend_Mode;

typedef enum { CAPY_LC_BUTT, CAPY_LC_ROUND, CAPY_LC_PROJECTION } CapyPDF_Line_Cap;

typedef enum {
    CAPY_LJ_MITER,
    CAPY_LJ_ROUND,
    CAPY_LJ_BEVEL,
} CapyPDF_Line_Join;

typedef enum {
    CAPY_DC_PAGE,
    CAPY_DC_COLOR_TILING,
    CAPY_DC_UCOLORED_TILING,
    CAPY_DC_FORM_XOBJECT,
    CAPY_DC_TRANSPARENCY_GROUP,
} CapyPDF_Draw_Context_Type;

typedef enum {
    CAPY_TR_SPLIT,
    CAPY_TR_BLINDS,
    CAPY_TR_BOX,
    CAPY_TR_WIPE,
    CAPY_TR_DISSOLVE,
    CAPY_TR_GLITTER,
    CAPY_TR_R,
    CAPY_TR_FLY,
    CAPY_TR_PUSH,
    CAPY_TR_COVER,
    CAPY_TR_UNCOVER,
    CAPY_TR_FADE,
} CapyPDF_Transition_Type;

typedef enum {
    CAPY_BOX_MEDIA,
    CAPY_BOX_CROP,
    CAPY_BOX_BLEED,
    CAPY_BOX_TRIM,
    CAPY_BOX_ART,
} CapyPDF_Page_Box;

typedef enum {
    CAPY_PDFX_1_2001,
    CAPY_PDFX_1A_2001,
    CAPY_PDFX_1A_2003,
    CAPY_PDFX_3_2002,
    CAPY_PDFX_3_2003,
    CAPY_PDFX_4,
    CAPY_PDFX_4P,
    CAPY_PDFX_5G,
    CAPY_PDFX_5PG,
} CapyPDF_PDFX_Type;

typedef enum {
    CAPY_INTERPOLATION_AUTO,
    CAPY_INTERPOLATION_PIXELATED,
    CAPY_INTERPOLATION_SMOOTH,
} CapyPDF_Image_Interpolation;

typedef enum {
    CAPY_COMPRESSION_NONE,
    CAPY_COMPRESSION_DEFLATE,
    // CAPY_COMPRESSION_CCIT4,
} CapyPDF_Compression;

typedef enum {
    CAPY_ANNOTATION_FLAG_NONE = 0,
    CAPY_ANNOTATION_FLAG_INVISIBLE = 1,
    CAPY_ANNOTATION_FLAG_HIDDEN = (1 << 1),
    CAPY_ANNOTATION_FLAG_PRINT = (1 << 2),
    CAPY_ANNOTATION_FLAG_NOZOOM = (1 << 3),
    CAPY_ANNOTATION_FLAG_NOROTATE = (1 << 4),
    CAPY_ANNOTATION_FLAG_NOVIEW = (1 << 5),
    CAPY_ANNOTATION_FLAG_READONLY = (1 << 6),
    CAPY_ANNOTATION_FLAG_LOCKED = (1 << 7),
    CAPY_ANNOTATION_FLAG_TOGGLENOVIEW = (1 << 8),
    CAPY_ANNOTATION_FLAG_LOCKEDCONTENTS = (1 << 9),
} CapyPDF_AnnotationFlags;

typedef enum {
    CAPY_STRUCTURE_TYPE_DOCUMENT,
    CAPY_STRUCTURE_TYPE_DOCUMENT_FRAGMENT,

    CAPY_STRUCTURE_TYPE_PART,
    CAPY_STRUCTURE_TYPE_SECT,
    CAPY_STRUCTURE_TYPE_DIV,
    CAPY_STRUCTURE_TYPE_ASIDE,
    CAPY_STRUCTURE_TYPE_NONSTRUCT,

    CAPY_STRUCTURE_TYPE_P,
    CAPY_STRUCTURE_TYPE_H,
    CAPY_STRUCTURE_TYPE_H1,
    CAPY_STRUCTURE_TYPE_H2,
    CAPY_STRUCTURE_TYPE_H3,
    CAPY_STRUCTURE_TYPE_H4,
    CAPY_STRUCTURE_TYPE_H5,
    CAPY_STRUCTURE_TYPE_H6,
    CAPY_STRUCTURE_TYPE_H7,
    CAPY_STRUCTURE_TYPE_TITLE,
    CAPY_STRUCTURE_TYPE_FENOTE,

    CAPY_STRUCTURE_TYPE_SUB,

    CAPY_STRUCTURE_TYPE_LBL,
    CAPY_STRUCTURE_TYPE_SPAN,
    CAPY_STRUCTURE_TYPE_EM,
    CAPY_STRUCTURE_TYPE_STRONG,
    CAPY_STRUCTURE_TYPE_LINK,
    CAPY_STRUCTURE_TYPE_ANNOT,
    CAPY_STRUCTURE_TYPE_FORM,

    CAPY_STRUCTURE_TYPE_RUBY,
    CAPY_STRUCTURE_TYPE_RB,
    CAPY_STRUCTURE_TYPE_RT,
    CAPY_STRUCTURE_TYPE_RP,
    CAPY_STRUCTURE_TYPE_WARICHU,
    CAPY_STRUCTURE_TYPE_WT,
    CAPY_STRUCTURE_TYPE_WP,

    CAPY_STRUCTURE_TYPE_L,
    CAPY_STRUCTURE_TYPE_LI,
    CAPY_STRUCTURE_TYPE_LBODY,
    CAPY_STRUCTURE_TYPE_TABLE,
    CAPY_STRUCTURE_TYPE_TR,
    CAPY_STRUCTURE_TYPE_TH,
    CAPY_STRUCTURE_TYPE_TD,
    CAPY_STRUCTURE_TYPE_THEAD,
    CAPY_STRUCTURE_TYPE_TBODY,
    CAPY_STRUCTURE_TYPE_TFOOT,

    CAPY_STRUCTURE_TYPE_CAPTION,

    CAPY_STRUCTURE_TYPE_FIGURE,

    CAPY_STRUCTURE_TYPE_FORMULA,

    CAPY_STRUCTURE_TYPE_ARTIFACT,

    CAPY_STRUCTURE_TYPE_NUM_ITEMS
} CapyPDF_StructureType;

// Note that some of these flags share the same bit values.
// That is what the specification says.
typedef enum {
    CAPY_FFIELD_NONE = 0,
    CAPY_FFIELD_READONLY = (1 << 0),
    CAPY_FFIELD_REQUIRED = (1 << 1),
    CAPY_FFIELD_NOEXPORT = (1 << 2),
    CAPY_FFIELD_MULTILINE = (1 << 12),
    CAPY_FFIELD_PASSWORD = (1 << 13),
    CAPY_FFIELD_NOTOGGLETOOFF = (1 << 14),
    CAPY_FFIELD_RADIO = (1 << 15),
    CAPY_FFIELD_PUSHBUTTON = (1 << 16),
    CAPY_FFIELD_COMBO = (1 << 17),
    CAPY_FFIELD_EDIT = (1 << 18),
    CAPY_FFIELD_SORT = (1 << 19),
    CAPY_FFIELD_FILESELECT = (1 << 20),
    CAPY_FFIELD_MULTISELECT = (1 << 21),
    CAPY_FFIELD_DONOTSPELLCHECK = (1 << 22),
    CAPY_FFIELD_DONOTSCROLL = (1 << 23),
    CAPY_FFIELD_COMB = (1 << 24),
    CAPY_FFIELD_RICHTEXT = (1 << 25),
    CAPY_FFIELD_RADIOSINUNUSON = (1 << 25),
    CAPY_FFIELD_COMMITONSELCHANGE = (1 << 26),
} CapyPDF_FormFieldFlags;

typedef enum {
    CAPY_FORM_ACTION_NONE = 0,
    CAPY_FORM_ACTION_INCLUEEXCLUDE = (1 << 0),
    CAPY_FORM_ACTION_INCLUDENOVALUEFIELDS = (1 << 1),
    CAPY_FORM_ACTION_EXPORTFORMAT = (1 << 2),
    CAPY_FORM_ACTION_GETMETHOD = (1 << 3),
    CAPY_FORM_ACTION_SUBMITCOORDINATES = (1 << 4),
    CAPY_FORM_ACTION_XFDF = (1 << 5),
    CAPY_FORM_ACTION_INCLUDEAPPENDSAVES = (1 << 6),
    CAPY_FORM_ACTION_INCLUDEANNOTATIONS = (1 << 7),
    CAPY_FORM_ACTION_SUBMITPDF = (1 << 8),
    CAPY_FORM_ACTION_CANONICALFORMAT = (1 << 9),
    CAPY_FORM_ACTION_EXCLNONUSERANNOTS = (1 << 10),
    CAPY_FORM_ACTION_EXCLFKEY = (1 << 11),
    CAPY_FORM_ACTION_EMBEDFORM = (1 << 12),
} CapyPDF_FormSubmitActionFlags;

// These start with a lower case letter, because all
// symbol names that start with an underscore + capital letter
// are reserved for the standard library.
typedef struct _capyPDF_Color CapyPDF_Color;
typedef struct _capyPDF_DrawContext CapyPDF_DrawContext;
typedef struct _capyPDF_Generator CapyPDF_Generator;
typedef struct _capyPDF_GraphicsState CapyPDF_GraphicsState;
typedef struct _capyPDF_OptionalContentGroup CapyPDF_OptionalContentGroup;
typedef struct _capyPDF_Options CapyPDF_Options;
typedef struct _capyPDF_PageProperties CapyPDF_PageProperties;
typedef struct _capyPDF_Text CapyPDF_Text;
typedef struct _capyPDF_TextSequence CapyPDF_TextSequence;
typedef struct _capyPDF_Transition CapyPDF_Transition;
typedef struct _capyPDF_RasterImage CapyPDF_RasterImage;
typedef struct _capyPDF_RasterImageBuilder CapyPDF_RasterImageBuilder;
typedef struct _capyPDF_Type2Function CapyPDF_Type2Function;
typedef struct _capyPDF_Type2Shading CapyPDF_Type2Shading;
typedef struct _capyPDF_Type3Shading CapyPDF_Type3Shading;
typedef struct _capyPDF_Type4Shading CapyPDF_Type4Shading;
typedef struct _capyPDF_Type6Shading CapyPDF_Type6Shading;
typedef struct _capyPDF_Annotation CapyPDF_Annotation;
typedef struct _capyPDF_StructItemExtraData CapyPDF_StructItemExtraData;
typedef struct _capyPDF_ImagePdfProperties CapyPDF_ImagePdfProperties;
typedef struct _capyPDF_Destination CapyPDF_Destination;
typedef struct _capyPDF_Outline CapyPDF_Outline;

typedef int32_t CapyPDF_EC;

typedef struct {
    int32_t id;
} CapyPDF_AnnotationId;

typedef struct {
    int32_t id;
} CapyPDF_EmbeddedFileId;

typedef struct {
    int32_t id;
} CapyPDF_FontId;

typedef struct {
    int32_t id;
} CapyPDF_FormWidgetId;

typedef struct {
    int32_t id;
} CapyPDF_FormXObjectId;

typedef struct {
    int32_t id;
} CapyPDF_GraphicsStateId;

typedef struct {
    int32_t id;
} CapyPDF_IccColorSpaceId;

typedef struct {
    int32_t id;
} CapyPDF_LabColorSpaceId;

typedef struct {
    int32_t id;
} CapyPDF_ImageId;

typedef struct {
    int32_t id;
} CapyPDF_OptionalContentGroupId;

typedef struct {
    int32_t id;
} CapyPDF_StructureItemId;

typedef struct {
    int32_t id;
} CapyPDF_TransparencyGroupId;

typedef struct {
    int32_t id;
} CapyPDF_FunctionId;

typedef struct {
    int32_t id;
} CapyPDF_ShadingId;

typedef struct {
    int32_t id;
} CapyPDF_OutlineId;

typedef struct {
    int32_t id;
} CapyPDF_SeparationId;

typedef struct {
    int32_t id;
} CapyPDF_PatternId;

typedef struct {
    int32_t id;
} CapyPDF_RoleId;

// Options

CAPYPDF_PUBLIC CapyPDF_EC capy_options_new(CapyPDF_Options **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_destroy(CapyPDF_Options *) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_title(CapyPDF_Options *opt,
                                                 const char *utf8_title) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_author(CapyPDF_Options *opt,
                                                  const char *utf8_author) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_creator(CapyPDF_Options *opt,
                                                   const char *utf8_creator) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_language(CapyPDF_Options *opt,
                                                    const char *lang) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_device_profile(
    CapyPDF_Options *opt, CapyPDF_DeviceColorspace cs, const char *profile_path) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_colorspace(CapyPDF_Options *opt,
                                                      CapyPDF_DeviceColorspace cs) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_output_intent(CapyPDF_Options *opt,
                                                         const char *identifier) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_pdfx(CapyPDF_Options *opt,
                                                CapyPDF_PDFX_Type xtype) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_default_page_properties(
    CapyPDF_Options *opt, const CapyPDF_PageProperties *prop) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_options_set_tagged(CapyPDF_Options *opt,
                                                  int32_t is_tagged) CAPYPDF_NOEXCEPT;

// Page properties.
CAPYPDF_PUBLIC CapyPDF_EC capy_page_properties_new(CapyPDF_PageProperties **out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_page_properties_destroy(CapyPDF_PageProperties *) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_page_properties_set_pagebox(CapyPDF_PageProperties *prop,
                                                           CapyPDF_Page_Box boxtype,
                                                           double x1,
                                                           double y1,
                                                           double x2,
                                                           double y2) CAPYPDF_NOEXCEPT;

// Generator

CAPYPDF_PUBLIC CapyPDF_EC capy_generator_new(const char *filename,
                                             const CapyPDF_Options *options,
                                             CapyPDF_Generator **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_page(CapyPDF_Generator *g,
                                                  CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_form_xobject(CapyPDF_Generator *gen,
                                                          CapyPDF_DrawContext *ctx,
                                                          CapyPDF_FormXObjectId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_color_pattern(
    CapyPDF_Generator *g, CapyPDF_DrawContext *ctx, CapyPDF_PatternId *out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_embed_jpg(CapyPDF_Generator *gen,
                                                   const char *fname,
                                                   CapyPDF_Image_Interpolation interpolate,
                                                   CapyPDF_ImageId *out_ptr) CAPYPDF_NOEXCEPT;
// FIXME, specify whether to compress the file or not.
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_embed_file(
    CapyPDF_Generator *g, const char *fname, CapyPDF_EmbeddedFileId *out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_load_font(CapyPDF_Generator *gen,
                                                   const char *fname,
                                                   CapyPDF_FontId *out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_load_image(CapyPDF_Generator *gen,
                                                    const char *fname,
                                                    CapyPDF_RasterImage **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_convert_image(CapyPDF_Generator *gen,
                                                       const CapyPDF_RasterImage *source,
                                                       CapyPDF_DeviceColorspace output_cs,
                                                       CapyPDF_Rendering_Intent ri,
                                                       CapyPDF_RasterImage **out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_image(CapyPDF_Generator *gen,
                                                   CapyPDF_RasterImage *image,
                                                   const CapyPDF_ImagePdfProperties *params,
                                                   CapyPDF_ImageId *out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_type2_function(CapyPDF_Generator *gen,
                                                            CapyPDF_Type2Function *func,
                                                            CapyPDF_FunctionId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_type2_shading(CapyPDF_Generator *gen,
                                                           CapyPDF_Type2Shading *shade,
                                                           CapyPDF_ShadingId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_type3_shading(CapyPDF_Generator *gen,
                                                           CapyPDF_Type3Shading *shade,
                                                           CapyPDF_ShadingId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_type4_shading(CapyPDF_Generator *gen,
                                                           CapyPDF_Type4Shading *shade,
                                                           CapyPDF_ShadingId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_type6_shading(CapyPDF_Generator *gen,
                                                           CapyPDF_Type6Shading *shade,
                                                           CapyPDF_ShadingId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_graphics_state(CapyPDF_Generator *gen,
                                                            const CapyPDF_GraphicsState *state,
                                                            CapyPDF_GraphicsStateId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_structure_item(CapyPDF_Generator *gen,
                                                            const CapyPDF_StructureType stype,
                                                            const CapyPDF_StructureItemId *parent,
                                                            CapyPDF_StructItemExtraData *extra,
                                                            CapyPDF_StructureItemId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC
capy_generator_add_custom_structure_item(CapyPDF_Generator *gen,
                                         const CapyPDF_RoleId role,
                                         const CapyPDF_StructureItemId *parent,
                                         CapyPDF_StructItemExtraData *extra,
                                         CapyPDF_StructureItemId *out_ptr) CAPYPDF_NOEXCEPT;

CAPYPDF_PUBLIC CapyPDF_EC capy_generator_load_icc_profile(
    CapyPDF_Generator *gen, const char *fname, CapyPDF_IccColorSpaceId *out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_lab_colorspace(CapyPDF_Generator *gen,
                                                            double xw,
                                                            double yw,
                                                            double zw,
                                                            double amin,
                                                            double amax,
                                                            double bmin,
                                                            double bmax,
                                                            CapyPDF_LabColorSpaceId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_create_separation_simple(CapyPDF_Generator *gen,
                                                                  const char *separation_name,
                                                                  const CapyPDF_Color *c,
                                                                  CapyPDF_SeparationId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_write(CapyPDF_Generator *gen) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC
capy_generator_add_optional_content_group(CapyPDF_Generator *gen,
                                          const CapyPDF_OptionalContentGroup *ocg,
                                          CapyPDF_OptionalContentGroupId *out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_create_annotation(CapyPDF_Generator *gen,
                                                           CapyPDF_Annotation *annotation,
                                                           CapyPDF_AnnotationId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_rolemap_entry(CapyPDF_Generator *gen,
                                                           const char *name,
                                                           CapyPDF_StructureType builtin,
                                                           CapyPDF_RoleId *out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_destroy(CapyPDF_Generator *gen) CAPYPDF_NOEXCEPT;

CAPYPDF_PUBLIC CapyPDF_EC capy_generator_add_outline(CapyPDF_Generator *gen,
                                                     const CapyPDF_Outline *outline,
                                                     CapyPDF_OutlineId *out_ptr) CAPYPDF_NOEXCEPT;

// Probably not final API.
CAPYPDF_PUBLIC CapyPDF_EC capy_generator_text_width(CapyPDF_Generator *gen,
                                                    const char *utf8_text,
                                                    CapyPDF_FontId font,
                                                    double pointsize,
                                                    double *out_ptr) CAPYPDF_NOEXCEPT;

// Draw context

CAPYPDF_PUBLIC CapyPDF_EC
capy_page_draw_context_new(CapyPDF_Generator *gen, CapyPDF_DrawContext **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_b(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_B(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_bstar(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_Bstar(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_BDC_builtin(
    CapyPDF_DrawContext *ctx, CapyPDF_StructureItemId structid) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_BDC_ocg(
    CapyPDF_DrawContext *ctx, CapyPDF_OptionalContentGroupId ocgid) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_BMC(CapyPDF_DrawContext *ctx,
                                          const char *tag) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_c(CapyPDF_DrawContext *ctx,
                                        double x1,
                                        double y1,
                                        double x2,
                                        double y2,
                                        double x3,
                                        double y3) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_cm(CapyPDF_DrawContext *ctx,
                                         double m1,
                                         double m2,
                                         double m3,
                                         double m4,
                                         double m5,
                                         double m6) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_d(CapyPDF_DrawContext *ctx,
                                        double *dash_array,
                                        int32_t array_size,
                                        double phase) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_EMC(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_f(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_fstar(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_G(CapyPDF_DrawContext *ctx, double gray) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_g(CapyPDF_DrawContext *ctx, double gray) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_gs(CapyPDF_DrawContext *ctx,
                                         CapyPDF_GraphicsStateId gsid) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_h(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_i(CapyPDF_DrawContext *ctx, double flatness) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_j(CapyPDF_DrawContext *ctx,
                                        CapyPDF_Line_Join join_style) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_J(CapyPDF_DrawContext *ctx,
                                        CapyPDF_Line_Cap cap_style) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC
capy_dc_cmd_k(CapyPDF_DrawContext *ctx, double c, double m, double y, double k) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC
capy_dc_cmd_K(CapyPDF_DrawContext *ctx, double c, double m, double y, double k) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_l(CapyPDF_DrawContext *ctx,
                                        double x,
                                        double y) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_m(CapyPDF_DrawContext *ctx,
                                        double x,
                                        double y) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_M(CapyPDF_DrawContext *ctx,
                                        double miterlimit) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_n(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_q(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_Q(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC
capy_dc_cmd_re(CapyPDF_DrawContext *ctx, double x, double y, double w, double h) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_RG(CapyPDF_DrawContext *ctx, double r, double g, double b)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_rg(CapyPDF_DrawContext *ctx, double r, double g, double b)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_ri(CapyPDF_DrawContext *ctx,
                                         CapyPDF_Rendering_Intent ri) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_s(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_S(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_sh(CapyPDF_DrawContext *ctx,
                                         CapyPDF_ShadingId shid) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_v(
    CapyPDF_DrawContext *ctx, double x2, double y2, double x3, double y3) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_w(CapyPDF_DrawContext *ctx,
                                        double line_width) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_W(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_Wstar(CapyPDF_DrawContext *ctx) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_cmd_y(
    CapyPDF_DrawContext *ctx, double x1, double y1, double x3, double y3) CAPYPDF_NOEXCEPT;

CAPYPDF_PUBLIC CapyPDF_EC capy_dc_set_stroke(CapyPDF_DrawContext *ctx,
                                             CapyPDF_Color *c) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_set_nonstroke(CapyPDF_DrawContext *ctx,
                                                CapyPDF_Color *c) CAPYPDF_NOEXCEPT;

CAPYPDF_PUBLIC CapyPDF_EC capy_dc_draw_image(CapyPDF_DrawContext *ctx,
                                             CapyPDF_ImageId iid) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_render_text(CapyPDF_DrawContext *ctx,
                                              const char *text,
                                              CapyPDF_FontId fid,
                                              double point_size,
                                              double x,
                                              double y) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_render_text_obj(CapyPDF_DrawContext *ctx,
                                                  CapyPDF_Text *text) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_set_page_transition(
    CapyPDF_DrawContext *ctx, CapyPDF_Transition *transition) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC
capy_dc_add_simple_navigation(CapyPDF_DrawContext *ctx,
                              const CapyPDF_OptionalContentGroupId *ocgarray,
                              int32_t array_size,
                              const CapyPDF_Transition *tr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_text_new(CapyPDF_DrawContext *ctx,
                                           CapyPDF_Text **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_set_custom_page_properties(
    CapyPDF_DrawContext *ctx, const CapyPDF_PageProperties *custom_properties);
CAPYPDF_PUBLIC CapyPDF_EC capy_dc_annotate(CapyPDF_DrawContext *ctx,
                                           CapyPDF_AnnotationId aid) CAPYPDF_NOEXCEPT;

CAPYPDF_PUBLIC CapyPDF_EC capy_dc_destroy(CapyPDF_DrawContext *) CAPYPDF_NOEXCEPT;

// Form XObjects.

CAPYPDF_PUBLIC CapyPDF_EC capy_form_xobject_new(CapyPDF_Generator *gen,
                                                double w,
                                                double h,
                                                CapyPDF_DrawContext **out_ptr) CAPYPDF_NOEXCEPT;

// Color Pattern

CAPYPDF_PUBLIC CapyPDF_EC capy_color_pattern_context_new(CapyPDF_Generator *gen,
                                                         CapyPDF_DrawContext **out_ptr,
                                                         double w,
                                                         double h) CAPYPDF_NOEXCEPT;

// Text

CAPYPDF_PUBLIC CapyPDF_EC capy_text_sequence_new(CapyPDF_TextSequence **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_sequence_append_codepoint(CapyPDF_TextSequence *tseq,
                                                              uint32_t codepoint) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_sequence_append_kerning(CapyPDF_TextSequence *tseq,
                                                            double kern) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_sequence_append_actualtext_start(
    CapyPDF_TextSequence *tseq, const char *actual_text) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_sequence_append_actualtext_end(CapyPDF_TextSequence *tseq)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_sequence_append_raw_glyph(CapyPDF_TextSequence *tseq,
                                                              uint32_t glyph_id,
                                                              uint32_t codepoint) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_sequence_destroy(CapyPDF_TextSequence *tseq) CAPYPDF_NOEXCEPT;

// This is basically cmd_Tj except that it does kerning if the font and Freetype support it.
CAPYPDF_PUBLIC CapyPDF_EC capy_text_render_text(CapyPDF_Text *text,
                                                const char *utf8_text) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_set_nonstroke(CapyPDF_Text *text,
                                                  const CapyPDF_Color *color) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_set_stroke(CapyPDF_Text *text,
                                               const CapyPDF_Color *color) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_BDC_builtin(CapyPDF_Text *text,
                                                    CapyPDF_StructureItemId stid) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_EMC(CapyPDF_Text *text) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_Tc(CapyPDF_Text *text, double spacing) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_Td(CapyPDF_Text *text, double x, double y) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_Tf(CapyPDF_Text *text,
                                           CapyPDF_FontId font,
                                           double pointsize) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_TJ(CapyPDF_Text *text,
                                           CapyPDF_TextSequence *kseq) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_TL(CapyPDF_Text *text, double leading) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_Tm(CapyPDF_Text *text,
                                           double a,
                                           double b,
                                           double c,
                                           double d,
                                           double e,
                                           double f) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_Tr(CapyPDF_Text *text,
                                           CapyPDF_Text_Mode tmode) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_Tw(CapyPDF_Text *text, double spacing) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_cmd_Tstar(CapyPDF_Text *text) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_text_destroy(CapyPDF_Text *text) CAPYPDF_NOEXCEPT;

// Color

CAPYPDF_PUBLIC CapyPDF_EC capy_color_new(CapyPDF_Color **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_color_destroy(CapyPDF_Color *color) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_color_set_rgb(CapyPDF_Color *c, double r, double g, double b)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_color_set_gray(CapyPDF_Color *c, double v) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC
capy_color_set_cmyk(CapyPDF_Color *color, double c, double m, double y, double k) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_color_set_icc(CapyPDF_Color *color,
                                             CapyPDF_IccColorSpaceId icc_id,
                                             double *values,
                                             int32_t num_values) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_color_set_separation(CapyPDF_Color *color,
                                                    CapyPDF_SeparationId sep_id,
                                                    double value) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_color_set_pattern(CapyPDF_Color *color,
                                                 CapyPDF_PatternId pat_id) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_color_set_lab(CapyPDF_Color *color,
                                             CapyPDF_LabColorSpaceId lab_id,
                                             double l,
                                             double a,
                                             double b) CAPYPDF_NOEXCEPT;

// Transition
CAPYPDF_PUBLIC CapyPDF_EC capy_transition_new(CapyPDF_Transition **out_ptr,
                                              CapyPDF_Transition_Type type,
                                              double duration) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_transition_destroy(CapyPDF_Transition *transition) CAPYPDF_NOEXCEPT;

// Optional Content groups
CAPYPDF_PUBLIC CapyPDF_EC capy_optional_content_group_new(CapyPDF_OptionalContentGroup **out_ptr,
                                                          const char *name) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_optional_content_group_destroy(CapyPDF_OptionalContentGroup *group)
    CAPYPDF_NOEXCEPT;

// Graphics state
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_new(CapyPDF_GraphicsState **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_set_CA(CapyPDF_GraphicsState *state,
                                                     double value) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_set_ca(CapyPDF_GraphicsState *state,
                                                     double value) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_set_BM(CapyPDF_GraphicsState *state,
                                                     CapyPDF_Blend_Mode blendmode) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_set_op(CapyPDF_GraphicsState *state,
                                                     int32_t value) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_set_OP(CapyPDF_GraphicsState *state,
                                                     int32_t value) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_set_OPM(CapyPDF_GraphicsState *state,
                                                      int32_t value) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_set_TK(CapyPDF_GraphicsState *state,
                                                     int32_t value) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_graphics_state_destroy(CapyPDF_GraphicsState *state)
    CAPYPDF_NOEXCEPT;

// Raster images
CAPYPDF_PUBLIC CapyPDF_EC capy_raster_image_builder_new(CapyPDF_RasterImageBuilder **out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_raster_image_builder_set_size(CapyPDF_RasterImageBuilder *builder,
                                                             int32_t w,
                                                             int32_t h) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_raster_image_builder_set_pixel_data(
    CapyPDF_RasterImageBuilder *builder, const char *buf, int32_t bufsize) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_raster_image_builder_destroy(CapyPDF_RasterImageBuilder *builder)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_raster_image_builder_build(
    CapyPDF_RasterImageBuilder *builder, CapyPDF_RasterImage **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_raster_image_get_colorspace(
    const CapyPDF_RasterImage *image, CapyPDF_ImageColorspace *out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_raster_image_has_profile(const CapyPDF_RasterImage *image,
                                                        int32_t *out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_raster_image_destroy(CapyPDF_RasterImage *image) CAPYPDF_NOEXCEPT;

// Functions
CAPYPDF_PUBLIC CapyPDF_EC capy_type2_function_new(double *domain,
                                                  int32_t domain_size,
                                                  const CapyPDF_Color *c1,
                                                  const CapyPDF_Color *c2,
                                                  double n,
                                                  CapyPDF_Type2Function **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type2_function_destroy(CapyPDF_Type2Function *func) CAPYPDF_NOEXCEPT;

// Shadings
CAPYPDF_PUBLIC CapyPDF_EC capy_type2_shading_new(CapyPDF_DeviceColorspace cs,
                                                 double x0,
                                                 double y0,
                                                 double x1,
                                                 double y1,
                                                 CapyPDF_FunctionId func,
                                                 int32_t extend1,
                                                 int32_t extend2,
                                                 CapyPDF_Type2Shading **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type2_shading_destroy(CapyPDF_Type2Shading *shade) CAPYPDF_NOEXCEPT;

CAPYPDF_PUBLIC CapyPDF_EC capy_type3_shading_new(CapyPDF_DeviceColorspace cs,
                                                 double *coords,
                                                 CapyPDF_FunctionId func,
                                                 int32_t extend1,
                                                 int32_t extend2,
                                                 CapyPDF_Type3Shading **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type3_shading_destroy(CapyPDF_Type3Shading *shade) CAPYPDF_NOEXCEPT;

CAPYPDF_PUBLIC CapyPDF_EC capy_type4_shading_new(CapyPDF_DeviceColorspace cs,
                                                 double minx,
                                                 double miny,
                                                 double maxx,
                                                 double maxy,
                                                 CapyPDF_Type4Shading **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type4_shading_add_triangle(CapyPDF_Type4Shading *shade,
                                                          const double *coords,
                                                          const CapyPDF_Color **colors)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type4_shading_extend(CapyPDF_Type4Shading *shade,
                                                    int32_t flag,
                                                    const double *coords,
                                                    const CapyPDF_Color *color) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type4_shading_destroy(CapyPDF_Type4Shading *shade) CAPYPDF_NOEXCEPT;

CAPYPDF_PUBLIC CapyPDF_EC capy_type6_shading_new(CapyPDF_DeviceColorspace cs,
                                                 double minx,
                                                 double miny,
                                                 double maxx,
                                                 double maxy,
                                                 CapyPDF_Type6Shading **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type6_shading_add_patch(CapyPDF_Type6Shading *shade,
                                                       const double *coords,
                                                       const CapyPDF_Color **colors)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type6_shading_extend(CapyPDF_Type6Shading *shade,
                                                    int32_t flag,
                                                    const double *coords,
                                                    const CapyPDF_Color **colors) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_type6_shading_destroy(CapyPDF_Type6Shading *shade) CAPYPDF_NOEXCEPT;

// Annotations.
CAPYPDF_PUBLIC CapyPDF_EC capy_text_annotation_new(const char *utf8_text,
                                                   CapyPDF_Annotation **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_file_attachment_annotation_new(
    CapyPDF_EmbeddedFileId fid, CapyPDF_Annotation **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_printers_mark_annotation_new(
    CapyPDF_FormXObjectId fid, CapyPDF_Annotation **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_annotation_set_rectangle(
    CapyPDF_Annotation *annotation, double x1, double y1, double x2, double y2) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_annotation_set_flags(CapyPDF_Annotation *annotation,
                                                    CapyPDF_AnnotationFlags flags) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_annotation_destroy(CapyPDF_Annotation *annotation) CAPYPDF_NOEXCEPT;

// Structural item extra data
CAPYPDF_PUBLIC CapyPDF_EC capy_struct_item_extra_data_new(CapyPDF_StructItemExtraData **out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_struct_item_extra_data_set_t(CapyPDF_StructItemExtraData *extra,
                                                            const char *title) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_struct_item_extra_data_set_lang(CapyPDF_StructItemExtraData *extra,
                                                               const char *lang) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_struct_item_extra_data_set_alt(CapyPDF_StructItemExtraData *extra,
                                                              const char *alt) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_struct_item_extra_data_set_actual_text(
    CapyPDF_StructItemExtraData *extra, const char *actual) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_struct_item_extra_data_destroy(CapyPDF_StructItemExtraData *extra)
    CAPYPDF_NOEXCEPT;

// Image load
CAPYPDF_PUBLIC CapyPDF_EC capy_image_pdf_properties_new(CapyPDF_ImagePdfProperties **out_ptr)
    CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_image_pdf_properties_set_mask(CapyPDF_ImagePdfProperties *par,
                                                             int32_t as_mask) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_image_pdf_properties_set_interpolate(
    CapyPDF_ImagePdfProperties *par, CapyPDF_Image_Interpolation interp) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_image_pdf_properties_set_conversion_intent(
    CapyPDF_ImagePdfProperties *par, CapyPDF_Rendering_Intent ri) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_image_pdf_properties_destroy(CapyPDF_ImagePdfProperties *par)
    CAPYPDF_NOEXCEPT;

// Destinations.
CAPYPDF_PUBLIC CapyPDF_EC capy_destination_new(CapyPDF_Destination **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_destination_set_page(CapyPDF_Destination *dest,
                                                    int32_t physical_page_number) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_destination_set_xyz(CapyPDF_Destination *dest,
                                                   double *x,
                                                   double *y,
                                                   double *z) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_destination_destroy(CapyPDF_Destination *dest) CAPYPDF_NOEXCEPT;

// Outline
CAPYPDF_PUBLIC CapyPDF_EC capy_outline_new(CapyPDF_Outline **out_ptr) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_outline_set_title(CapyPDF_Outline *outline,
                                                 const char *title) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_outline_set_destination(
    CapyPDF_Outline *outline, const CapyPDF_Destination *dest) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_outline_set_rgb(CapyPDF_Outline *outline,
                                               double r,
                                               double g,
                                               double b) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_outline_set_f(CapyPDF_Outline *outline, uint32_t F) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_outline_set_parent(CapyPDF_Outline *outline,
                                                  CapyPDF_OutlineId parent) CAPYPDF_NOEXCEPT;
CAPYPDF_PUBLIC CapyPDF_EC capy_outline_destroy(CapyPDF_Outline *outline) CAPYPDF_NOEXCEPT;

// Error

CAPYPDF_PUBLIC const char *capy_error_message(CapyPDF_EC error_code) CAPYPDF_NOEXCEPT;

#ifdef __cplusplus
}
#endif
